/*----------------------------------------------------------------------
PROGRAM: convolve.c
DATE:    5/5/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------

  2D filtering routines

---------------------------------------------------------------------- */

#include "util.h"





/* ---------------------------------------------------------- */
/*
   This conv2d_sep routine takes a float image of size (M1,M2),
   a separable float filter of size (N1,N2), and computes their
   PERIODIC convolution.

   NOTE: since this implements periodic convolution the resulting
         image is the same size as the input (ie. M1 by M2)
*/
/* ---------------------------------------------------------- */

void conv2d_sep(float **image, float **result, int M1, int M2,
		float *filter1, int N1, float *filter2, int N2)
{
  register int i,j,k,l;
  int left,center,offset,shift,m;
  float **temp,sum;

  center = (int) 1 + N2/2;  /* center-tap of 1d filter2 */
  left = center - 1;
  offset = (int) left/M2 + 1;   /* offset to add to handle negative indices */
  offset *= M2;                 /* in modulus-M2 call                       */
  shift = N2 - left - 1 + offset;        
                                      /* (N2-k+1) flips filter    */
                                      /* k-1, j-1 and k-left give */
                                      /* zero-origin indices      */

  /* filter the rows first using filter2[1...N2] */

  temp = matrix(1,M1,1,M2);
  for(i=1;i<=M1;i++) 
    for(j=1;j<=M2;j++) { 
      m = j + shift;
      for(k=1,sum=0;k<=N2;k++) {
        l = (m - k)%M2;   
        l++;
        sum += filter2[k] * image[i][l];
      }
     temp[i][j] = sum; 
    }


  /* filter columns of temp using filter1[1...N1] */

  center = (int) 1 + N1/2;  /* center-tap of 1d filter1 */
  left = center - 1;
  offset = (int) left/M1 + 1;   /* offset to add to handle negative indices */
  offset *= M1;                 /* in modulus-M1 call                       */ 
  shift = N1 - left - 1 + offset;
                                      /* (N1-k+1) flips filter    */
                                      /* k-1, j-1 and k-left give */
                                      /* zero-origin indices      */

  for(j=1;j<=M2;j++) 
    for(i=1;i<=M1;i++) {
      m = i + shift;
      for(k=1,sum=0;k<=N1;k++) {
        l = (m - k)%M1;
        l++;
        sum += filter1[k] * temp[l][j];
      }
     result[i][j] = sum;
    }

  free_matrix(temp,1,M1,1,M2);

}

/* ---------------------------------------------------------------------- */


/* ---------------------------------------------------------- */
/*
   This conv2d routine takes a float image of size (M1,M2),
   a NON-separable float filter of size (N1,N2), and computes their
   PERIODIC convolution.

   NOTE: since this implements periodic convolution the resulting
         image is the same size as the input (ie. M1 by M2)
*/
/* ---------------------------------------------------------- */

void conv2d(float **image, float **result, int M1, int M2,
	    float **filter, int N1, int N2)
{
  register int i,j,k,l;
  int left1,center1,offset1,shift1,m1,i1,j1;
  int left2,center2,offset2,shift2,m2,i2,j2;
  float sum;


  center1 = (int) 1 + N1/2;  /* center-tap of 1d filter1 */
  left1 = center1 - 1;
  offset1 = (int) left1/M1 + 1; /* offset to add to handle negative indices */
  offset1 *= M1;                /* in modulus-M1 call                       */ 
  shift1 = N1 - left1 - 1 + offset1;
                                      /* (N1-k+1) flips filter    */
                                      /* k-1, j-1 and k-left give */
                                      /* zero-origin indices      */

  center2 = (int) 1 + N2/2;  /* center-tap of 1d filter2 */
  left2 = center2 - 1;
  offset2 = (int) left2/M2 + 1; /* offset to add to handle negative indices */
  offset2 *= M2;                /* in modulus-M2 call                       */
  shift2 = N2 - left2 - 1 + offset2;        
                                      /* (N2-k+1) flips filter    */
                                      /* k-1, j-1 and k-left give */
                                      /* zero-origin indices      */


  /* filter the input image */

  for(i=1;i<=M1;i++) {
    m1 = i + shift1;
    for(j=1;j<=M2;j++) { 
      m2 = j + shift2;
      sum = 0;
      for(k=1;k<=N1;k++) {
	i1 = (m1 - k)%M1;   
        i1++;
	for(l=1;l<=N2;l++) {
	  i2 = (m2 - l)%M2;   
	  i2++;
	  sum += filter[k][l] * image[i1][i2];
	}
      }
      result[i][j] = sum; 
    }
  }

/* -----------------------------
  for(j=1;j<=M2;j++) 
    for(i=1;i<=M1;i++) {
      m = i + shift;
      for(k=1,sum=0;k<=N1;k++) {
        l = (m - k)%M1;
        l++;
        sum += filter1[k] * temp[l][j];
      }
     result[i][j] = sum;
    }


  for(i=1;i<=M1;i++) 
    for(j=1;j<=M2;j++) { 
      m = j + shift;
      for(k=1,sum=0;k<=N2;k++) {
        l = (m - k)%M2;   
        l++;
        sum += filter2[k] * image[i][l];
      }
     temp[i][j] = sum; 
    }
---------------------------------- */

}

/* ---------------------------------------------------------- */
/*
   This conv2d_sep_point routine takes a float image of size (M1,M2),
   a separable float filter of size (N1,N2), and computes the
   resulting value of the output at coordinates (x,y) 

   NOTE: This routine computes the periodic convolution!
*/
/* ---------------------------------------------------------- */

float conv2d_sep_point(float **image, int M1, int M2, int x, int y,
		       float *filter1, int N1, float *filter2, int N2)
{
  register int i,j,k,l;
  int i1,left1,center1,offset1,shift1,m1;
  int left2,center2,offset2,shift2,m2;
  float *temp,result,sum;

  center2= (int) 1 + N2/2;  /* center-tap of 1d filter2 */
  left2 = center2 - 1;
  offset2 = (int) left2/M2 + 1;   /* offset to add to handle negative indices */
  offset2 *= M2;                 /* in modulus-M2 call                       */
  shift2 = N2 - left2 - 1 + offset2;        

  center1 = (int) 1 + N1/2;  /* center-tap of 1d filter2 */
  left1 = center1 - 1;
  offset1 = (int) left1/M1 + 1;   /* offset to add to handle negative indices */
  offset1 *= M1;                 /* in modulus-M2 call                       */
  shift1 = N1 - left1 - 1 + offset1;


  /* filter the rows first using filter2[1...N2] */

  temp = vector(1,N1);
  m2 = y + shift2;
  i1 = x - left1 - 1;
  for(j=1;j<=N1;j++) {
     i  = (i1 + j -1 + offset1)%M1;
     i++;
     for(k=1,sum=0;k<=N2;k++) {
       l = (m2 - k)%M2;   
       l++;
       sum += filter2[k] * image[i][l];
     }
     temp[j] = sum; 
  }

  /* filter point the column at (x,y) using filter1[1...N1] */

  for(i=1,result=0;i<=N1;i++) 
   result += filter1[N1-i+1] * temp[i];

  free_vector(temp,1,N1);
  return result;
}

